@@ -36,6 +36,7 @@ typedef struct snd_pcm_pulse {
     /* Since ALSA expects a ring buffer we must do some voodoo. */
     size_t last_size;
     size_t ptr;
+    int underrun;
 
     size_t offset;
 
@@ -57,7 +58,7 @@ static void update_ptr(snd_pcm_pulse_t *pcm)
 
     if (size > pcm->last_size) {
         pcm->ptr += size - pcm->last_size;
-        pcm->ptr %= pcm->buffer_attr.maxlength;
+        pcm->ptr %= pcm->buffer_attr.tlength;
     }
 
     pcm->last_size = size;
@@ -90,7 +91,9 @@ static int pulse_start(snd_pcm_ioplug_t *io)
     if (err < 0) {
         err = -EIO;
         goto finish;
-    }
+    } else
+    	pcm->underrun = 0;
+
 
 finish:
     pa_threaded_mainloop_unlock(pcm->p->mainloop);
@@ -200,6 +203,9 @@ static snd_pcm_sframes_t pulse_pointer(snd_pcm_ioplug_t *io)
 
 	err = snd_pcm_bytes_to_frames(io->pcm, pcm->ptr);
 
+	if (pcm->underrun)
+		err = -EPIPE;
+
 finish:
     pa_threaded_mainloop_unlock(pcm->p->mainloop);
 
@@ -231,6 +237,9 @@ static int pulse_delay(snd_pcm_ioplug_t *io,
 
 	*delayp = snd_pcm_bytes_to_frames(io->pcm, pa_usec_to_bytes(lat, &pcm->ss));
 
+ 	if (pcm->underrun && pcm->io.state == SND_PCM_STATE_RUNNING)
+		snd_pcm_ioplug_set_state(io, SND_PCM_STATE_XRUN);
+
 finish:
     pa_threaded_mainloop_unlock(pcm->p->mainloop);
 
@@ -273,6 +282,7 @@ static snd_pcm_sframes_t pulse_write(snd_pcm_ioplug_t *io,
         pulse_poll_deactivate(pcm->p);
 
     err = size;
+    pcm->underrun = 0;
 
 finish:
     pa_threaded_mainloop_unlock(pcm->p->mainloop);
@@ -354,6 +364,15 @@ static void stream_request_cb(pa_stream *p, size_t length, void *userdata)
     pulse_poll_activate(pcm->p);
 }
 
+static void stream_underrun_cb(pa_stream *p, void *userdata) {
+	snd_pcm_pulse_t *pcm = userdata;
+
+	assert(pcm);
+	assert(pcm->p);
+
+	pcm->underrun = 1;
+}
+
 static int pulse_pcm_poll_descriptors_count(snd_pcm_ioplug_t *io)
 {
 	snd_pcm_pulse_t *pcm = io->private_data;
@@ -461,6 +480,7 @@ static int pulse_prepare(snd_pcm_ioplug_t *io)
 
     if (io->stream == SND_PCM_STREAM_PLAYBACK) {
         pa_stream_set_write_callback(pcm->stream, stream_request_cb, pcm);
+	pa_stream_set_underflow_callback(pcm->stream, stream_underrun_cb, pcm);
         pa_stream_connect_playback(pcm->stream, pcm->device, &pcm->buffer_attr,
         	PA_STREAM_AUTO_TIMING_UPDATE | PA_STREAM_INTERPOLATE_TIMING, NULL, NULL);
     } else {
@@ -480,6 +500,7 @@ static int pulse_prepare(snd_pcm_ioplug_t *io)
     pcm->last_size = 0;
     pcm->ptr = 0;
     pcm->offset = 0;
+    pcm->underrun = 0;
 
 finish:
     pa_threaded_mainloop_unlock(pcm->p->mainloop);
@@ -490,15 +511,26 @@ finish:
 static int pulse_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params)
 {
     snd_pcm_pulse_t *pcm = io->private_data;
+    snd_pcm_t *base = io->pcm;
 	int err = 0;
 
     assert(pcm);
     assert(pcm->p);
 
+    //Resolving bugtrack ID 0003470
+    if (!base) {
+	switch (snd_pcm_state(base)) {
+	case SND_PCM_STATE_SETUP:
+	case SND_PCM_STATE_PREPARED:
+	    break;
+	default:
+    	    assert(!pcm->stream);
+	    break;
+	}
+    }
+    
     pa_threaded_mainloop_lock(pcm->p->mainloop);
 
-    assert(!pcm->stream);
-
     pcm->frame_size = (snd_pcm_format_physical_width(io->format) * io->channels) / 8;
 
     switch (io->format) {
@@ -533,9 +565,9 @@ static int pulse_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params)
     pcm->ss.rate = io->rate;
     pcm->ss.channels = io->channels;
 
-    pcm->buffer_attr.maxlength = io->buffer_size * pcm->frame_size;
+    pcm->buffer_attr.maxlength = (io->buffer_size*3)/2 * pcm->frame_size;
     pcm->buffer_attr.tlength = io->buffer_size * pcm->frame_size;
-    pcm->buffer_attr.prebuf = io->period_size * pcm->frame_size;
+    pcm->buffer_attr.prebuf = (io->buffer_size - io->period_size) * pcm->frame_size;
     pcm->buffer_attr.minreq = io->period_size * pcm->frame_size;
     pcm->buffer_attr.fragsize = io->period_size * pcm->frame_size;
 
@@ -643,26 +675,18 @@ static int pulse_hw_constraint(snd_pcm_pulse_t *pcm)
     if (err < 0)
         return err;
 
-    err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_RATE,
-        8000, 48000);
-    if (err < 0)
-        return err;
-
-    err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_PERIOD_BYTES,
-        1, 4294967295U);
-    if (err < 0)
-        return err;
-
-    err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_PERIODS,
-        2, 4294967295U);
-    if (err < 0)
-        return err;
-
     err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_BUFFER_BYTES,
-        1, 4294967295U);
+        1, 4 * 1024 * 1024);
     if (err < 0)
         return err;
 
+	err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_PERIOD_BYTES, 256, 1024 * 1024);
+	if (err < 0)
+		return err;
+
+	err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_PERIODS, 2, 1024);
+	if (err < 0)
+		return err;
     return 0;
 }
 
